/*
 * kgraft_patch_bsc1122233
 *
 * Fix for regression bsc#1122233
 *
 *  Upstream commit:
 *  none, SUSE specific
 *
 *  SLE12(-SP1) commit:
 *  not affected
 *
 *  SLE12-SP2 commit:
 *  not affected
 *
 *  SLE12-SP3 commit:
 *  afe02740e899b82243c686ca10f0f41964ed4af8
 *
 *  SLE12-SP4 commit
 *  not affected
 *
 *  SLE15 commit:
 *  not affected
 *
 *
 *  Copyright (c) 2019 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_HP_WATCHDOG)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/watchdog.h>
#include <linux/device.h>
#include <linux/io.h>
#include "kgr_patch_bsc1122233.h"
#include "kallsyms_relocs.h"

#if !IS_MODULE(CONFIG_HP_WATCHDOG)
#error "Live patch supports only CONFIG_HP_WATCHDOG=m"
#endif

#define KGR_PATCHED_MODULE "hpwdt"


static unsigned long __iomem * *kgr_hpwdt_timer_reg;


static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "hpwdt_timer_reg", (void *)&kgr_hpwdt_timer_reg, "hpwdt" },
};



/* from drivers/watchdog/hpwdt.c */
#define KGR_SECS_TO_TICKS(secs)		((secs) * 1000 / 128)



/* patched, inlined */
int kgr_hpwdt_ping(struct watchdog_device *wdd)
{
	/*
	 * Fix bsc#1122233
	 *  +2 lines
	 */
	int reload = KGR_SECS_TO_TICKS(wdd->timeout);

	iowrite16(reload, (*kgr_hpwdt_timer_reg));

	return 0;
}

/* patched, calls inlined hpwdt_ping() */
int kgr_hpwdt_settimeout(struct watchdog_device *wdd, unsigned int val)
{
	wdd->timeout = val;
	kgr_hpwdt_ping(wdd);

	return 0;
}



static int kgr_patch_bsc1122233_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1122233_module_nb = {
	.notifier_call = kgr_patch_bsc1122233_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1122233_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1122233_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1122233_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1122233_module_nb);
}

#endif /* IS_ENABLED(CONFIG_HP_WATCHDOG) */
